#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include "ijs.h"
#include "ijs_server.h"

#define BUF_SIZE 4096

typedef struct _ExampleParamList ExampleParamList;

struct _ExampleParamList {
  ExampleParamList *next;
  char *key;
  char *value;
  int value_size;
};

static int
example_set_cb (void *set_cb_data, IjsServerCtx *ctx,
		const char *key, const char *value, int value_size)
{
  ExampleParamList **ppl = (ExampleParamList **)set_cb_data;
  ExampleParamList *pl = (ExampleParamList *)malloc (sizeof (ExampleParamList));
  int key_len = strlen (key);

  fprintf (stderr, "example_set_cb: %s=", key);
  fwrite (value, 1, value_size, stderr);
  fputs ("\n", stderr);
  pl->next = *ppl;
  pl->key = malloc (key_len + 1);
  memcpy (pl->key, key, key_len + 1);
  pl->value = malloc (value_size);
  memcpy (pl->value, value, value_size);
  pl->value_size = value_size;
  *ppl = pl;
  return 0;
}

/**
 * Finds a parameter in the param list, and allocates a null terminated
 * string with the value.
 **/
static char *
find_param (ExampleParamList *pl, const char *key)
{
  ExampleParamList *curs;
  char *result;

  for (curs = pl; curs != NULL; curs = curs->next)
    {
      if (!strcmp (curs->key, key))
	{
	  result = malloc (curs->value_size + 1);
	  memcpy (result, curs->value, curs->value_size);
	  result[curs->value_size] = 0;
	  return result;
	}
    }
  return NULL;
}

static void
free_param_list (ExampleParamList *pl)
{
  ExampleParamList *next;

  for (; pl != NULL; pl = next)
    {
      free (pl->key);
      free (pl->value);
      free (pl);
    }
}

int
main (int argc, char **argv)
{
  IjsServerCtx *ctx;
  IjsPageHeader ph;
  int status;
  char buf[BUF_SIZE];
  char hexbuf[BUF_SIZE * 3];
  char *fn;
  FILE *f = NULL;
  double xscale, yscale;
  ExampleParamList *pl = NULL;

  ctx = ijs_server_init ();
  ijs_server_install_set_cb (ctx, example_set_cb, &pl);
  if (ctx == NULL)
    exit (1);

  do 
    {
      int total_bytes, bytes_left;
      ExampleParamList *curs;

      status = ijs_server_get_page_header (ctx, &ph);
      if (status) break;
      fprintf (stderr, "got page header, %d x %d\n",
	      ph.width, ph.height);

      if (f == NULL)
	{
	  fn = find_param (pl, "OutputFile");
	  /* todo: check error! */

	  if (fn != NULL)
	    f = fopen (fn, "w");
	  if (f == NULL)
	    {
	      fprintf (stderr, "can't open output file %s\n", fn);
	      fclose (stdin);
	      fclose (stdout);
	      break;
	    }
	}

      fprintf (f, "%%!PS-Adobe-2.0\n"
	       "/rhex { currentfile exch readhexstring pop } bind def\n");
      fprintf (f, "/picstr %d string def\n", ph.width);

      for (curs = pl; curs != NULL; curs = curs->next)
	{
	  fprintf (f, "%% IJS parameter: %s = ", curs->key);
	  fwrite (curs->value, 1, curs->value_size, f);
	  fputs ("\n", f);
	}

      xscale = 72.0 / ph.xres;
      yscale = 72.0 / ph.yres;
      fprintf (f,
	       "gsave\n"
	       "0 0 translate\n"
	       "%f %f scale\n"
	       "%d %d %d\n"
	       "[ %d 0 0 %d 0 %d ]\n",
	       xscale * ph.width, yscale * ph.height,
	       ph.width, ph.height, ph.bps,
	       ph.width, -ph.height, ph.height);
      if (ph.n_chan == 1)
	fprintf (f, "{ picstr rhex } image\n");
      else
	{
	  fprintf (f, "{ picstr rhex }\n"
		   "false 3 colorimage\n");
	}
      total_bytes = ((ph.n_chan * ph.bps * ph.width + 7) >> 3) * ph.height;
      bytes_left = total_bytes;
      while (bytes_left)
	{
	  int n_bytes = bytes_left;
	  int i, j;

	  if (n_bytes > sizeof(buf))
	    n_bytes = sizeof(buf);
#ifdef VERBOSE
	  fprintf (stderr, "%d bytes left, reading %d\n", bytes_left, n_bytes);
#endif
	  status = ijs_server_get_data (ctx, buf, n_bytes);
	  if (status)
	    {
	      fprintf (stderr, "page aborted!\n");
	      break;
	    }
	  j = 0;
	  for (i = 0; i < n_bytes; i++)
	    {
	      const char hex[16] = "0123456789AbCdEf";
	      unsigned char c = ((unsigned char *)buf)[i];

	      hexbuf[j++] = hex[c >> 4];
	      hexbuf[j++] = hex[c & 0xf];
	      if ((i & 31) == 31)
		hexbuf[j++] = '\n';
	    }
	  if ((n_bytes & 31) != 0)
	    hexbuf[j++] = '\n';
	  fwrite (hexbuf, 1, j, f);
	  bytes_left -= n_bytes;
	}
      fprintf (f, "grestore\nshowpage\n");
    }
  while (status == 0);

  if (status > 0) status = 0; /* normal exit */

  if (f)
    fclose (f);
  ijs_server_done (ctx);

  free_param_list (pl);

#ifdef VERBOSE
  fprintf (stderr, "server exiting with status %d\n", status);
#endif
  return status;
}
